class SemanticError(SyntaxError):
    pass


class EntryError(NameError):
    pass


class ConfigImportError(ImportError):
    pass


class ExceptionWrapper(Exception):
    def __init__(self, exception):
        self.exception = exception



import os
from collections import OrderedDict, Counter
from pathlib import Path
from typing import Union, Dict, Any, Sequence

from .semantics import Semantics
from .exceptions import EntryError, ExceptionWrapper, SemanticError
from .scope import Scope, Builtins
from .parser import parse_file, parse_string
from .render import ScopeEval
from .statements import ImportConfig, Definitions

PathLike = Union[Path, str]


class Config:
    """
    A config interpreter.

    Parameters
    ----------
    shortcuts: dict, optional
        a dict that maps keywords to paths. It is used to resolve paths during import.
    injections: dict, optional
        a dict with default values that will be used in case the config doesn't define them.
    """
    # restricting setattr to these names
    __slots__ = '_shortcuts', '_imported_configs', '_builtins', '_scope', '_extension', '_injections'

    def __init__(self, shortcuts: Dict[str, PathLike] = None, injections: Dict[str, Any] = None):
        self._shortcuts = shortcuts or {}
        self._imported_configs = {}
        self._injections = injections
        self._builtins = Builtins(injections or {})
        self._scope: Scope = Scope([], self._builtins, {})
        self._extension = '.config'

    @classmethod
    def load(cls, path: PathLike, shortcuts: Dict[str, PathLike] = None, injections: Dict[str, Any] = None):
        """
        Import the config located at `path` and return a Config instance.
        Also this method adds a `__file__ = pathlib.Path(path)` value to the global scope.

        Parameters
        ----------
        path: str, pathlib.Path
            path to the config to import
        shortcuts: dict, optional
            a dict that maps keywords to paths. It is used to resolve paths during import.
        injections: dict, optional
            a dict with default values that will be used in case the config doesn't define them.

        Returns
        -------
        config: Config
        """
        # TODO: __folder__
        key = '__file__'
        injections = dict(injections or {})
        if key in injections:
            raise ValueError(f'The "{key}" key is not allowed in "injections".')

        injections[key] = cls._standardize_path(path)
        return cls(shortcuts, injections).file_input(path)

    @classmethod
    def loads(cls, source: str, shortcuts: Dict[str, PathLike] = None, injections: Dict[str, Any] = None):
        """
        Interpret the `source` and return a `Config` instance.

        Parameters
        ----------
        source: str
        shortcuts: dict, optional
            a dict that maps keywords to paths. It is used to resolve paths during import.
        injections: dict, optional
            a dict with default values that will be used in case the config doesn't define them.

        Returns
        -------
        config: Config
        """
        return cls(shortcuts, injections).string_input(source)

    def dumps(self, entry_points: Union[Sequence[str], str] = None) -> str:
        """
        Generate a string containing the names from the current scope.

        Parameters
        ----------
        entry_points
            the definitions that should be kept (along with their dependencies).
            If None - all the definitions are rendered.
        """
        if isinstance(entry_points, str):
            entry_points = [entry_points]
        return '\n'.join(self._scope.render(entry_points)).strip() + '\n'

    def dump(self, path: PathLike, entry_points: Union[Sequence[str], str] = None):
        """ Render the config and save it to `path`. See `dumps` for details. """
        with open(path, 'w') as file:
            file.write(self.dumps(entry_points))

    def file_input(self, path: PathLike) -> 'Config':
        """Import the config located at `path`."""
        self._update_scope(self._import(path))
        return self

    def string_input(self, source: str) -> 'Config':
        """Interpret the `source`."""
        self._update_scope(self._make_scope(*parse_string(source, self._extension)))
        return self

    def update(self, **values: Any) -> 'Config':
        """Update the scope by `values`."""
        try:
            self._scope.update_values(values)
            return self
        except RuntimeError:
            raise RuntimeError(
                'The scope has already been populated with live objects. Overwriting them might cause '
                'undefined behaviour. Please, create another instance or copy of Config: config.copy().update(...)'
            ) from None

    def copy(self) -> 'Config':
        """Create a copy of the config with an unpopulated scope."""
        config = type(self)(self._shortcuts, self._injections)
        config._scope = Scope.copy(self._scope)
        return config

    def __contains__(self, name: str):
        return name in self._scope

    def __getattr__(self, name: str):
        try:
            return self.get(name)
        except EntryError:
            raise AttributeError(f'"{name}" is not defined.')  # from None

    def keys(self):
        return self._scope.keys()

    def __iter__(self):
        yield from self.keys()

    def __getitem__(self, name: str):
        try:
            return self.get(name)
        except EntryError:
            raise KeyError(f'"{name}" is not defined.') from None

    def get(self, name: str):
        try:
            return self._scope[name]
        except ExceptionWrapper as e:
            raise e.exception from None

    def eval(self, expression: str):
        """Evaluate the given `expression`."""
        try:
            return eval(expression, ScopeEval(self._scope))
        except ExceptionWrapper as e:
            raise e.exception from None

    def _update_scope(self, scope: OrderedDict):
        self._scope.check_populated()
        # update
        new_scope = self._scope.definitions.copy()
        new_scope.update(scope)
        definitions = list(new_scope.values())
        # analysis
        tree = Semantics(definitions, self._builtins)
        tree.check()
        # building
        self._scope = Scope(definitions, self._builtins, tree.parents)

    @staticmethod
    def _standardize_path(path: PathLike) -> str:
        path = str(path)
        path = os.path.expanduser(path)
        path = os.path.realpath(path)
        return path

    def _import(self, path: str) -> OrderedDict:
        path = self._standardize_path(path)

        if path in self._imported_configs:
            return self._imported_configs[path]
        # avoiding cycles
        self._imported_configs[path] = {}

        result = self._make_scope(*parse_file(path, self._extension))
        self._imported_configs[path] = result
        return result

    def _make_scope(self, parents: Sequence[ImportConfig], definitions: Definitions) -> OrderedDict:
        parent_scope = OrderedDict()
        for parent in parents:
            parent_scope.update(self._import(parent.get_path(self._shortcuts)))

        duplicates = [name for name, count in Counter(x.name for x in definitions).items() if count > 1]
        if duplicates:
            source_path = definitions[0].statement.source_path
            duplicates = ', '.join(duplicates)
            raise SemanticError(f'Duplicate definitions found in {source_path}:\n    {duplicates}')

        final_scope = OrderedDict(parent_scope.items())
        final_scope.update((x.name, x) for x in definitions)
        return final_scope

    def __dir__(self):
        return list(set(self._scope.keys()) | set(super().__dir__()))

    def _ipython_key_completions_(self):
        return list(self._scope.keys())

    def __setattr__(self, name, value):
        try:
            super().__setattr__(name, value)
        except AttributeError:
            raise AttributeError(f'Config\'s attribute "{name}" is read-only.') from None


load = Config.load
loads = Config.loads



from threading import Lock
from typing import Any

from .statements import GlobalStatement, LiveObject


class Thunk:
    value: Any
    ready: bool


class ValueThunk(Thunk):
    def __init__(self, statement: LiveObject):
        self.statement = statement
        self.value = statement.value
        self.ready = True


class NodeThunk(Thunk):
    def __init__(self, statement: GlobalStatement, lock: Lock):
        self.lock = lock
        self.statement = statement
        self.ready = False
        self._value = None

    def set(self, value):
        assert not self.ready
        self._value = value
        self.ready = True

    @property
    def value(self):
        assert self.ready
        return self._value



import ast
import sys
from typing import Dict, Any

from .exceptions import ExceptionWrapper, EntryError


class ScopeWrapper(Dict[str, Any]):
    def __init__(self, scope):
        super().__init__()
        self.scope = scope

    def __getitem__(self, name):
        try:
            return self.scope[name]
        except KeyError as e:
            # this is needed because KeyError is converted to NameError by `eval/exec`
            raise ExceptionWrapper(e) from e
        except EntryError:
            pass

        if name not in self:
            raise NameError(f'The name "{name}" is not defined.')
        return super().__getitem__(name)

    def keys(self):
        return list(set(super().keys()) | set(self.scope.keys()))

    def get(self, key, default=None):
        if key in self:
            return self[key]
        return default

    def __iter__(self):
        yield from self.keys()

    def __len__(self):
        return len(self.keys())

    def values(self):
        raise NotImplementedError

    def items(self):
        raise NotImplementedError

    def popitem(self):
        raise NotImplementedError

    def setdefault(self, *args, **kwargs):
        raise NotImplementedError

    def update(self, **kwargs):
        raise NotImplementedError

    def __delitem__(self, v):
        raise NotImplementedError


class ScopeEval(ScopeWrapper):
    def __contains__(self, name):
        return name in self.scope or super().__contains__(name)

    def __setitem__(self, k, v):
        raise NotImplementedError


class ScopeExec(ScopeWrapper):
    def __setitem__(self, name, value):
        assert not super().__contains__(name)
        super().__setitem__(name, value)

    def get_results(self, names):
        values = []
        for name in names:
            values.append(super().pop(name))
        return values

    def __contains__(self, name):
        return name in self.scope or super().__contains__(name)


def execute(statement, global_scope, source):
    exec(compile(make_module([statement]), source, 'exec'), global_scope)


if sys.version_info[:2] >= (3, 8):
    def make_module(body):
        # handling `type_ignores`
        return ast.Module(body, [])
else:
    def make_module(body):
        return ast.Module(body)



import ast

from ..visitor import Visitor


class SemanticVisitor(Visitor):
    """Simple visitor for nodes that don't interact with the scope stack."""

    # utils

    def _visit_valid(self, value):
        if value is not None:
            self.visit(value)

    def _ignore_node(self, node):
        pass

    # literals

    visit_constant = visit_name_constant = visit_ellipsis = visit_bytes = visit_num = visit_str = _ignore_node

    def visit_formatted_value(self, node):
        assert node.format_spec is None
        self.visit(node.value)

    def visit_joined_str(self, node):
        self._iterate_nodes(node.values)

    def visit_list(self, node: ast.List):
        assert isinstance(node.ctx, ast.Load)
        self._iterate_nodes(node.elts)

    visit_tuple = visit_list

    def visit_set(self, node):
        self._iterate_nodes(node.elts)

    def visit_dict(self, node):
        self._iterate_nodes(filter(None, node.keys))
        self._iterate_nodes(node.values)

    # variables

    def visit_starred(self, node: ast.Starred):
        self.visit(node.value)

    # expressions

    def visit_unary_op(self, node: ast.UnaryOp):
        self.visit(node.operand)

    def visit_bin_op(self, node: ast.BinOp):
        self.visit(node.left)
        self.visit(node.right)

    def visit_bool_op(self, node: ast.BoolOp):
        self._iterate_nodes(node.values)

    def visit_compare(self, node: ast.Compare):
        self.visit(node.left)
        self._iterate_nodes(node.comparators)

    def visit_call(self, node: ast.Call):
        self.visit(node.func)
        self._iterate_nodes(node.args)
        self._iterate_nodes(node.keywords)
        self._visit_valid(getattr(node, 'starargs', None))
        self._visit_valid(getattr(node, 'kwargs', None))

    def visit_keyword(self, node):
        self.visit(node.value)

    def visit_if_exp(self, node: ast.IfExp):
        self.visit(node.test)
        self.visit(node.body)
        self.visit(node.orelse)

    def visit_attribute(self, node: ast.Attribute):
        assert isinstance(node.ctx, ast.Load)
        self.visit(node.value)

    # subscripting

    def visit_subscript(self, node: ast.Subscript):
        assert isinstance(node.ctx, ast.Load)
        self.visit(node.value)
        self.visit(node.slice)

    def visit_index(self, node):
        self.visit(node.value)

    def visit_slice(self, node):
        self._visit_valid(node.lower)
        self._visit_valid(node.upper)
        self._visit_valid(node.step)

    def visit_ext_slice(self, node):
        self._iterate_nodes(node.dims)

    # statements

    def visit_assert(self, node):
        self.visit(node.test)
        self._visit_valid(node.msg)

    def visit_if(self, node: ast.If):
        self.visit(node.test)
        self._iterate_nodes(node.body)
        self._iterate_nodes(node.orelse)

    def visit_expr(self, node: ast.Expr):
        self.visit(node.value)

    # helpers

    def visit_arguments(self, node: ast.arguments):
        self._iterate_nodes(node.defaults)
        self._iterate_nodes(filter(None, node.kw_defaults))



import sys
import ast
import bisect
from io import BytesIO
from tokenize import tokenize
from typing import Tuple, Sequence

from .visitor import Visitor
from .statements import GlobalFunction, GlobalAssign, ImportConfig, GlobalImportFrom, GlobalImport, Definition, \
    IGNORE_NAME, Definitions

NO_DECORATORS = sys.version_info[:2] > (3, 7)


def throw(message, position):
    raise SyntaxError(message + '\n  at %d:%d in %s' % position)


def get_substring(lines: Sequence[str], start_line: int, start_col: int, stop_line: int = None,
                  stop_col: int = None, lstrip: bool = True, rstrip: bool = True, keep_line: bool = True) -> str:
    lines = list(lines[start_line - 1:stop_line])

    lines[-1] = lines[-1][:stop_col]
    lines[0] = lines[0][start_col:]
    empty = 0

    # remove comments
    if lstrip:
        line = lines[0].strip()
        while line.startswith('#') or not line:
            lines.pop(0)
            line = lines[0].strip()

    if rstrip:
        line = lines[-1].strip()
        while line.startswith('#') or not line:
            if not line:
                empty += 1

            lines.pop()
            line = lines[-1].strip()

    body = '\n'.join(lines).strip()
    if keep_line and empty > 1:
        body += '\n'

    return body


def tokenize_string(source):
    return tokenize(BytesIO(source.encode()).readline)


def get_position(node, source):
    return node.lineno, node.col_offset, source


def extract_assign_targets(targets, source):
    def _extract(target):
        assert isinstance(target.ctx, ast.Store)

        if isinstance(target, ast.Name):
            yield target.id
        elif isinstance(target, ast.Starred):
            yield from _extract(target.value)
        elif isinstance(target, (ast.Tuple, ast.List)):
            for elt in target.elts:
                yield from _extract(elt)

        else:
            throw('This assignment syntax is not supported', get_position(target, source))

    result = []
    for t in targets:
        result.extend(_extract(t))
    return result


class Normalizer(Visitor):
    def __init__(self, start, stop, lines, source_path, extension):
        self.source_path = source_path
        self.lines = lines
        self.start = start
        self.stop = stop
        self.extension = extension

    def get_position(self, node: ast.AST):
        return get_position(node, self.source_path)

    def generic_visit(self, node, *args, **kwargs):
        throw('This syntactic structure is not supported.', self.get_position(node))

    def visit_function_def(self, node: ast.FunctionDef):
        body = get_substring(self.lines, *self.start, *self.stop)
        yield Definition(node.name, GlobalFunction(node, body, self.get_position(node)))

    def visit_assign(self, node: ast.Assign):
        position = self.get_position(node.value)

        last_target = node.targets[-1]
        body = get_substring(self.lines, last_target.lineno, last_target.col_offset, *self.stop)
        # TODO: use a tokenizer
        body = body[body.index('=') + 1:].lstrip()

        statement = GlobalAssign(node, body, position)
        for target in node.targets:
            names = list(extract_assign_targets([target], self.source_path))
            if set(names) == {IGNORE_NAME}:
                throw('The assignment target cannot completely consist of the `_` wildcard', self.get_position(target))

            for name in names:
                if name != IGNORE_NAME:
                    yield Definition(name, statement)

    def visit_import_from(self, node: ast.ImportFrom):
        names = node.names
        root = node.module.split('.')
        position = self.get_position(node)
        # starred config import
        if len(names) == 1 and names[0].name == '*':
            yield ImportConfig(root, node.level, self.extension, position)
            return
        # relative imports make no sense, as there is no base module in configs
        if node.level > 0:
            throw('Relative imports are only supported for config files.', position)
        # absolute imports
        for alias in names:
            local = ast.copy_location(ast.ImportFrom(node.module, [alias], 0), node)
            statement = GlobalImportFrom(local, position)
            yield Definition(statement.name, statement)

    def visit_import(self, node: ast.Import):
        position = self.get_position(node)

        for alias in node.names:
            local = ast.copy_location(ast.Import([alias]), node)
            statement = GlobalImport(local, position)
            yield Definition(statement.name, statement)


# need this function, because in >=3.8 the function start is considered from `def` token
#   rather then from the first decorator
def find_body_limits(source: str, source_path: str):
    def _pos(node):
        return node.lineno, node.col_offset

    statements = sorted(ast.parse(source, source_path).body, key=_pos, reverse=True)
    tokens = list(tokenize_string(source))
    if not tokens:
        return

    indices = [t.start for t in tokens]
    stop = tokens[-1].end

    for statement in statements:
        start = _pos(statement)
        if NO_DECORATORS and isinstance(statement, ast.FunctionDef) and statement.decorator_list:
            dec = statement.decorator_list[0]
            start = _pos(dec)
            idx = bisect.bisect_left(indices, start)
            token = tokens[idx]
            assert token.start == start, (token, start)
            token = tokens[idx - 1]
            assert token.string == '@', token
            start = token.start

        yield statement, start, stop
        stop = start


def parse(source: str, source_path: str, extension: str) -> Tuple[Sequence[ImportConfig], Definitions]:
    lines = tuple(source.splitlines() + [''])
    wrapped = []
    for statement, start, stop in reversed(list(find_body_limits(source, source_path))):
        wrapped.extend(Normalizer(start, stop, lines, source_path, extension).visit(statement))

    parents, imports, definitions = [], [], []
    # TODO: move this to normalizer?
    for entry in wrapped:
        if isinstance(entry, ImportConfig):
            if imports or definitions:
                throw('Starred imports are only allowed at the top of the config.', entry.position)
            parents.append(entry)

        elif isinstance(entry.statement, GlobalImport):
            if definitions:
                throw('Imports are only allowed before definitions.', entry.statement.position)
            imports.append(entry)

        else:
            assert isinstance(entry.statement, (GlobalAssign, GlobalFunction))
            definitions.append(entry)

    return parents, imports + definitions


def parse_file(config_path, extension):
    with open(config_path, 'r') as file:
        return parse(file.read(), config_path, extension)


def parse_string(source, extension):
    return parse(source, '<string input>', extension)



from distutils.core import setup
from setuptools import find_packages
from lazycon import __version__

classifiers = '''Development Status :: 5 - Production/Stable
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9'''

with open('README.md', encoding='utf-8') as file:
    long_description = file.read()

setup(
    name='lazycon',
    packages=find_packages(include=('lazycon',)),
    include_package_data=True,
    version=__version__,
    description='Easy config files in pure Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='maxme1',
    author_email='maxs987@gmail.com',
    license='MIT',
    url='https://github.com/maxme1/lazycon',
    download_url='https://github.com/maxme1/lazycon/archive/v%s.tar.gz' % __version__,
    keywords=[
        'config', 'lazy', 'interpreter'
    ],
    classifiers=classifiers.splitlines(),
    install_requires=[],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'lazycon = lazycon.console:main',
        ],
    },
)
